# contains class for flashcards with question and answers
from pygame.sprite import Sprite
from pygame.font import SysFont
from pygame.image import load
from pygame import Rect, Surface


class Card(Sprite):
    def __init__(self, x: int = 110, y: int = 120):
        """:param x starting x-position
        :param y starting y-position"""
        super().__init__()
        self.x, self.y = x, y
        self.font, self.text_sprite, self.text_rect = None, None, None
        self.question, self.answer = None, None
        self.rect, self.image = None, None
        self.front, self.back = None, None
        self.flipped = False

    def set_fields(self, question: str, answer: str):
        """Set question & answer fields, used in addition to __init__ for instantiating a flashcard
        :param question text of the card
        :param answer text for the card"""
        self.question, self.answer = question, answer
        return self

    def set_pos(self, x: int, y: int):
        """Set the card's position
        :param x new x-position
        :param y new y-position"""
        self.x, self.y = x, y

    def set_imgs(self, front: str, back: str):
        """Load the images initially
        :param front image path for the self.flipped=False side
        :param back image path for the self.flipped=True side"""
        self.front, self.back = front, back
        self.refresh_img()

    def load_text(self, font):
        """Load the text font and text sprite (without rendering)"""
        self.font = SysFont(font, 30, italic=False, bold=False)
        self.text_sprite = self.font.render(self.question, True, (255, 255, 255))
        self.text_rect = self.text_sprite.get_rect()

    def render_text(self, surface, x_offset=250, y_offset=115):
        """Renders text without wrap"""
        if self.flipped:
            txt = self.answer
        else:
            txt = self.question
        self.text_rect.center = (self.x + x_offset, self.y + y_offset)
        self.text_sprite = self.font.render(txt, True, (255, 255, 255))
        surface.blit(self.text_sprite, self.text_rect)

    def render_textwrap(self, surface: Surface, color: (int, int, int) = None,
                        font: SysFont = None, padding_left: int = 40, rect: Rect = None, aa=True):
        """Renders text on a card while wrapping anything that goes over its rect boundaries
           Function adapted from https://www.pygame.org/wiki/TextWrap
           :param surface which screen to draw text on
           :param color text color
           :param font text font
           :param padding_left how much space to leave on the left
           :param rect the rect to draw text on
           :param aa enable anti-aliasing?"""
        if color is None:
            color = (255, 255, 255)
        if rect is None:
            rect = self.rect
        if self.flipped:
            text = self.answer
        else:
            text = self.question
        if font is None:
            font = self.font

        lineSpacing = -2
        # get the height of the font
        fontHeight = font.size("Tg")[1]
        max_lines = int((rect.height - fontHeight) / (fontHeight + lineSpacing))

        # calculate the total number of lines for the given text
        rectWidth = rect.width - padding_left
        words = text.split()
        # calculate the width of each word
        wordWidths = [font.size(word)[0] for word in words]
        # calculate the number of lines required
        numLines = 1
        lineWidth = 0
        for i, word in enumerate(words):
            # add the width of the current word to the line width
            lineWidth += wordWidths[i]
            # add the width of the space character
            if i < len(words) - 1:
                lineWidth += font.size(' ')[0]
            # if the line width exceeds the width of the rect, start a new line
            if lineWidth > rectWidth:
                numLines += 1
                lineWidth = wordWidths[i]

        # center the text on the y-axis
        y = rect.top + (rect.height - (numLines * (fontHeight + lineSpacing) - lineSpacing)) // 2

        while text:
            i = 1
            # determine if the row of text will be outside our area
            if y + fontHeight > rect.bottom:
                break
            # determine maximum width of line
            while font.size(text[:i])[0] < rect.width - padding_left and i < len(text):
                i += 1
            # if we've wrapped the text, then adjust the wrap to the last word
            if i < len(text):
                i = text.rfind(" ", 0, i) + 1
            image = font.render(text[:i], aa, color)
            surface.blit(image, (rect.left + padding_left, y))
            y += fontHeight + lineSpacing
            # remove the text we just blitted
            text = text[i:]
        return text

    def refresh_img(self):
        """Load the image corresponding to what side self.flipped shows"""
        if self.flipped:
            self.image = load(self.back)
        else:
            self.image = load(self.front)
        self.rect = self.image.get_rect()

    def flip(self):
        """Invert self.flipped and the card image"""
        self.flipped = not self.flipped
        self.refresh_img()

    def update(self) -> None:
        """Override pygame's default sprite update() to set the card's rect (actual) position to the correct values"""
        self.rect.x, self.rect.y = self.x, self.y




from pygame.font import SysFont


class Text:
    """Manages all text to display onscreen (except for flashcards text)"""
    def __init__(self, text: str, x: int, y: int, font: str, size: int = 30, red: int = 255, green: int = 255, blue: int = 255):
        self.text = text
        self.field = SysFont(font, size, italic=False, bold=False)
        self.font = font
        self.sprite = self.field.render(self.text, True, (red, green, blue))
        self.rect = self.sprite.get_rect()
        self.x, self.y = x, y

    def render_text(self, surface, x_offset=0, y_offset=0):
        """Renders text without wrap"""
        self.rect.center = (self.x + x_offset, self.y + y_offset)
        self.field = self.field.render(self.text, True, (255, 255, 255))
        surface.blit(self.sprite, self.rect)



from .card import Card
from pygame.sprite import Group as SpriteGroup


class Deck:
    def __init__(self):
        """
        Attributes:
        self.cards = the deck, a group of cards
        self.card_in_focus = the first card of the deck, the one to show onscreen
        self.focused_card_number = the card number of the focused card based on the initial deck order
        """
        self.cards = SpriteGroup()
        self.card_in_focus = SpriteGroup()
        self.focused_card_number = 1

    def add_card(self, question: str, answer: str) -> 'Deck':
        """:param question the question for the Card to add to the deck
            :param answer the Card's answer"""
        self.cards.add(Card().set_fields(question, answer))
        return self

    def add_card_from_list(self, add: [str]) -> 'Deck':
        """:param add list of two strings formatted [question, answer]"""
        self.cards.add(Card().set_fields(add[0], add[1]))
        return self

    def add_card_from_obj(self, add: Card):
        """:param add the Card object to add to this deck"""
        self.cards.add(add)
        return self

    def init_cards(self, front: str, back: str, font: str) -> 'Deck':
        """Set the background images for all self.cards
        :param front .jpg (500x250) pixels to use for the front
        :param back .jpg to use for the back
        :param font name of the pygame font to load"""
        for card in self.cards:
            card.set_imgs(front, back)
            card.load_text(font)
        return self

    def get_first_card(self) -> SpriteGroup:
        """Get the current focused card as a pygame.Sprite.Group"""
        temp_group = SpriteGroup()
        temp_group.add(self.cards.sprites()[0])
        return temp_group

    def move_to_back(self, mv_card: Card) -> SpriteGroup:
        """Move the given card to the back of self.cards
        :param mv_card    the card to move to the back"""
        self.unflip()
        self.cards.remove(mv_card)
        self.cards.add(mv_card)
        return self.cards

    def move_first_to_back(self) -> SpriteGroup:
        """Move the self.card_in_focus to the back of self.cards"""
        self.iterate_focused_card(1)
        return self.move_to_back(self.cards.sprites()[0])

    def move_to_front(self, mv_card: Card) -> SpriteGroup:
        """Move the given card to the front of the deck
        :param mv_card   the card to move to the front"""
        self.unflip()
        grouplist = []
        for c in self.cards:
            grouplist.append(c)
        grouplist.remove(mv_card)
        grouplist.insert(0, mv_card)
        tempgroup = SpriteGroup()
        for c in grouplist:
            tempgroup.add(c)
        self.cards = tempgroup
        return self.cards

    def unflip(self):
        """Reset all cards in the deck to the flipped=False side"""
        for card in self.cards:
            if card.flipped:
                card.flip()

    def move_last_to_front(self) -> SpriteGroup:
        """Move the last card in the deck to the front"""
        self.iterate_focused_card(-1)
        return self.move_to_front(self.cards.sprites()[-1])

    def iterate_focused_card(self, by: int) -> int:
        """Iterates the self.focused_card_number variable, and resets it to 1 or len(self.cards) if needed
        :param by  the number to add to the current card number"""
        if self.focused_card_number + by > len(self.cards):
            self.focused_card_number = 1
        elif self.focused_card_number + by < 1:
            self.focused_card_number = len(self.cards)
        else:
            self.focused_card_number += by
        return self.focused_card_number

    def print(self):
        """Print each question: answer in the deck"""
        for card in self.cards:
            print(f"{card.question}: {card.answer}")



from argparse import ArgumentParser
from .gui.constants import Screen
from os import path


def handle_args():
    """Create all arguments, and return the parsed given arguments"""
    args = ArgumentParser()
    pkg_dir = path.dirname(path.abspath(__file__))
    args.add_argument("--fps",
                      metavar="FPS",
                      type=int,
                      help="Frames per second to run at",
                      default=30),
    args.add_argument("--file",
                      metavar="INPUT_FILE",
                      help="Path to the file containing questions and answers for your flashcards",
                      default=path.join(pkg_dir, "default.csv"))
    args.add_argument("--font",
                      metavar="FONT_NAME",
                      type=str,
                      help="Name of the pygame-supported font to use",
                      default="arial")
    args.add_argument("--cards_front",
                      metavar="IMG_PATH",
                      help="Path to the .jpg to use for all flashcard front backgrounds (500x250 pixels)",
                      default=path.join(pkg_dir, "assets/card_front.jpg"))
    args.add_argument("--cards_back",
                      metavar="IMG_PATH",
                      help="Path to the .jpg to use for all flashcard back backgrounds (500x250 pixels)",
                      default=path.join(pkg_dir, "assets/card_back.jpg"))
    args.add_argument("--h",
                      type=int,
                      metavar="WIN_HEIGHT",
                      help="Window height",
                      default=Screen.height)
    args.add_argument("--w",
                      type=int,
                      metavar="WIN_WIDTH",
                      help="Window width",
                      default=Screen.width)
    args.add_argument("--title",
                      type=str,
                      metavar="WIN_TITLE",
                      help="Window title",
                      default=Screen.title)
    return args.parse_args()



import pygame
from ..flashcards import Deck
from pygame.sprite import Group as SpriteGroup


class GUI:
    def __init__(self, opts):
        self.screen, self.clock = None, None
        self.window_name = opts.title
        self.dimensions = (opts.w, opts.h)
        self.FPS = opts.fps

    def init_screen(self) -> 'GUI':
        """initialize the pygame environment, reset its clock, and initialize the pygame gui window"""
        pygame.init()
        self.reset_clock()
        self.screen = pygame.display.set_mode(self.dimensions)
        pygame.display.set_caption(self.window_name)
        return self

    def reset_clock(self):
        """reset pygame's clock"""
        self.clock = pygame.time.Clock()

    def render(self, group: SpriteGroup):
        """Render the card visible in the current frame
        :param group the group of cards to render (Deck().cards)"""
        self.clock.tick(self.FPS)
        card = group.sprites()[0]  # in this game we won't ever display anything except the first card in a deck
        card.update()
        card.render_textwrap(self.screen)
        pygame.display.update()

    @staticmethod
    def handle_events(deck: Deck) -> (bool, Deck):
        """Given a deck of flashcards, handle the 'pygame events' and determine the new deck orientation
            The user inputs will determine what card should be moved to the front/back, and the new deck is returned
            along with whether the game should continue running or not (based on if the quit button was pressed)
        :param deck the deck of flashcards to use, manipulate, and return"""
        for event in pygame.event.get():
            keys = pygame.key.get_pressed()
            if event.type == pygame.MOUSEBUTTONDOWN or keys[pygame.K_UP] or keys[pygame.K_DOWN]:
                deck.cards.sprites()[0].flip()  # first card of the deck is the only one in view
            if keys[pygame.K_RIGHT]:
                deck.move_first_to_back()
            if keys[pygame.K_LEFT]:
                deck.move_last_to_front()
            if keys[pygame.K_ESCAPE] or event.type == pygame.QUIT:
                return False, deck
        return True, deck

    @staticmethod
    def get_mouse_pos() -> (int, int):
        """return the mouse position according to pygame"""
        return pygame.mouse.get_pos()

    @staticmethod
    def quit():
        """quit pygame"""
        pygame.quit()



from .deck import Deck


# For parsing csv documents into question | answer format


def from_csv(filepath: str) -> Deck:
    """Takes the first two columns of the filepath csv and translates them into col1: questions col2: answers
    :param filepath the path to the input csv"""
    csv_deck = Deck()
    with open(filepath, "r") as f:
        for line in f.readlines():
            csv_deck.add_card_from_list(line.strip().split(","))
    return csv_deck



class Screen:
    """Sets the default argument values for
        --title
        --w
        --h
        """
    title = "FlashQuiz"
    width = 700
    height = 500
    dimensions = (width, height)



from setuptools import setup, find_packages


setup(
    name='flashquiz',
    version="0.4",
    description='Practice your flashcards using python',
    author='mike-fmh',
    author_email='mikemh@uri.edu',
    license='MIT',
    packages=find_packages(),
    package_data={
        'flashquiz': ['default.csv', 'assets/*']
    },
    entry_points={'console_scripts': ['flashquiz = flashquiz.quizzer:main']},
    install_requires=[
        'pygame',
        'argparse'
    ]
)



from flashquiz.flashcards.parser import from_csv
from flashquiz.args import handle_args
from flashquiz.gui import GUI, Text


def main():
    args = handle_args()
    game = GUI(args).init_screen()

    # convert the csv given in args to our deck
    deck = from_csv(args.file).init_cards(args.cards_front, args.cards_back, args.font)
    run = True
    while run:
        game.screen.fill((0, 0, 0))

        # create text that displays card number
        card_number_txt = Text(f"#{deck.focused_card_number}", 100, 100, args.font)
        card_number_txt.render_text(game.screen)

        # set card_in_focus to first card in the deck (we only need to render the first card)
        card_in_focus = deck.get_first_card()
        card_in_focus.draw(game.screen)
        game.render(card_in_focus)

        # handle_events() will determine any changes to make to the deck based on the user input
        # (what should be the new card in focus) and returns the updated deck
        # it also keeps track of if we should continue running based on if the quit key has been pressed (escape)
        run, deck = game.handle_events(deck)

    game.quit()


if __name__ == '__main__':
    main()
